"""
Projects data onto the SLM's virtual display, using the :mod:`pyglet` library.
"""
import os
import ctypes
import warnings
import numpy as np

from slmsuite.hardware.slms.slm import SLM

try:
    import pyglet
    import pyglet.gl as gl

    if pyglet.__version__[0] != "1":
        raise ImportError("pyglet >=2.0.0 is not supported. Please downgrade to 1.5.29.")
except ImportError:
    warnings.warn("pyglet not installed. Install to use ScreenMirrored SLMs.")

class ScreenMirrored(SLM):
    """
    Wraps a :mod:`pyglet` window for displaying data to an SLM.

    Important
    ~~~~~~~~~
    Many SLM manufacturers provide an SDK for interfacing with their devices.
    Using a python wrapper for these SDKs is recommended, instead of or in supplement to this class,
    as there often is functionality additional to a mirrored screen
    (e.g. USB for changing settings) along with device-specific optimizations.

    Note
    ~~~~
    There are a variety of python packages that support blitting images onto a fullscreen display.

    -   `Simple DirectMedia Layer (SDL) <https://www.libsdl.org/>`_ wrappers:

        - :mod:`pygame` (`link <https://www.pygame.org/docs/>`__),
          which also supports OpenGL. Only supports one screen.
        - :mod:`sdl2` (`readthedocs <https://pysdl2.readthedocs.io/en/latest/>`__)
          through the ``PySDL2`` package. Requires additional libraries.

    -   `Open Graphics Library (OpenGL) <https://www.opengl.org/>`_ wrappers:

        - :mod:`moderngl` (`readthedocs <https://moderngl.readthedocs.io/en/latest/>`__),
          an OpenGL wrapper focusing on a pythonic interface for core OpenGL functions.
        - :mod:`OpenGL` (`link <http://pyopengl.sourceforge.net/documentation/index.html>`__)
          through the ``PyOpenGL``/``PyOpenGL_accelerate`` package, a very light OpenGL wrapper.
        - :mod:`pyglet` (`readthedocs <https://pyglet.readthedocs.io/en/latest/>`__),
          a light OpenGL wrapper.

    -   GUI Library wrappers:

        - :mod:`gi` (`readthedocs <https://pygobject.readthedocs.io/en/latest/>`__),
          through the ``PyGObject`` package wrapping ``GTK`` and other GUI libraries.
        - :mod:`pyqt6` (`link <https://riverbankcomputing.com/software/pyqt/>`__),
          through the ``PyQt6`` package wrapping the version 6 ``Qt`` GUI library.
        - :mod:`tkinter` (`link <https://docs.python.org/3/library/tkinter.html>`__),
          included in standard ``python``, wrapping the ``Tcl``/``Tk`` GUI library.
        - :mod:`wx` (`link <https://docs.wxpython.org/>`__),
          through the ``wxPython`` package wrapping the ``wxWidgets`` GUI library.
          :mod:`slmpy` (`GitHub <https://github.com/wavefrontshaping/slmPy>`__) uses :mod:`wx`.

    :mod:`slmsuite` uses :mod:`pyglet` as the default display package.
    :mod:`pyglet` is generally more capable than the mentioned SDL wrappers while immediately supporting
    features such as detecting connected displays which low-level packages like :mod:`OpenGL` and
    :mod:`moderngl` do not have. :mod:`pyglet` allows us to interact more directly with the display
    hardware without the additional overhead that is found in GUI libraries.
    Most importantly, :mod:`pyglet` is well documented.

    However, it might be worthwhile in the future to look back into SDL options, as SDL surfaces
    are closer to the pixels than OpenGL textures, so greater speed might be achievable (even without
    loading data to the GPU as a texture). Another potential improvement could come from writing
    :mod:`cupy` datastructures to ``OpenGL`` textures directly, without using the CPU as an
    intermediary. There is `some precedent <https://docs.nvidia.com/cuda/cuda-runtime-api/group__CUDART__OPENGL.html>`_
    for transferring data from ``CUDA`` (on which :mod:`cupy` is based) to ``OpenGL``,
    though :mod:`cupy` does not currently directly support this.

    Important
    ~~~~~~~~~
    :class:`ScreenMirrored` uses a double-buffered and vertically synchronized (vsync) ``OpenGL``
    context. This is to prevent "tearing" resulting from data being modified during a display write:
    rather, all monitor writes are synchronized such that clean frames are always displayed.
    This feature is similar to the ``isImageLock`` flag in :mod:`slmpy`, but is implemented a bit
    closer to the hardware.

    Attributes
    ----------
    window : pyglet.window.Window
        Fullscreen window used to send information to the SLM.
    tex_shape_ratio : (int, int)
        Ratio between the SLM shape and the (power-2-padded) texture stored in ``OpenGL``.
    buffer : numpy.ndarray
        Memory used to load data to the ``OpenGL`` memory. Of type ``np.uint8``.
    cbuffer : pyglet.gl.GLubyte
        Array of length ``prod(shape) * 4`` (4 bytes per RGBA).
        Maps to the same memory as :attr:`buffer`.
        Used to load data to the texture.
    texture : pyglet.gl.GLuint
        Identifier for the texture loaded into ``OpenGL`` memory.
    """

    def __init__(
            self,
            display_number,
            bitdepth=8,
            wav_um=1,
            pitch_um=(8,8),
            verbose=True,
            **kwargs
        ):
        """
        Initializes a :mod:`pyglet` window for displaying data to an SLM.

        Caution
        ~~~~~~~
        An SLM designed at 1064 nm can be used for an application at 780 nm by passing
        ``wav_um=.780`` and ``wav_design_um=1.064``,
        thus causing the SLM to use only a fraction (780/1064)
        of the full dynamic range. Be sure these values are correct.
        Note that there are some performance losses from using this modality (see :meth:`.write()`).

        Caution
        ~~~~~~~
        There is some subtlety to
        `complex display setups with Linux <https://pyglet.readthedocs.io/en/latest/modules/canvas.html>`_.
        Working outside the default display is currently not implemented.

        Parameters
        ----------
        display_number : int
            Monitor number for frame to be instantiated upon.
        bitdepth : int
            Bitdepth of the SLM. Defaults to 8.

            Caution
            ~~~~~~~
            This class currently supports SLMs with 8-bit precision or less.
            In the future, this class will also support 16-bit SLMs using RG color.
        wav_um : float
            Wavelength of operation in microns. Defaults to 1 um.
        pitch_um : (float, float)
            Pixel pitch in microns. Defaults to 8 micron square pixels.
        verbose : bool
            Whether or not to print extra information.
        **kwargs
            See :meth:`.SLM.__init__` for permissible options.
        """
        if verbose:
            print("Initializing pyglet... ", end="")
        display = pyglet.canvas.get_display()
        screens = display.get_screens()
        if verbose:
            print("success")
            print("Searching for window with display_number={}... "
                    .format(display_number), end="")

        if len(screens) <= display_number:
            raise ValueError("Could not find display_number={}; only {} displays"
                .format(display_number, len(screens)))

        screen_info = ScreenMirrored.info(verbose=False)

        if screen_info[display_number][3]:
            raise ValueError(
                "ScreenMirrored window already created on display_number={}"
                .format(display_number))

        if verbose and screen_info[display_number][2]:
            print("warning: this is the main display... ")

        if verbose:
            print("success")
            print("Creating window... ", end="")

        screen = screens[display_number]

        super().__init__(
            (screen.width, screen.height),
            bitdepth=bitdepth,
            wav_um=wav_um,
            pitch_um=pitch_um,
            **kwargs
        )

        # Setup the window. If failure, closes elegantly upon except().
        try:
            # Make the window and do basic setup.
            self.window = pyglet.window.Window( screen=screen,
                                                fullscreen=True, vsync=True)
            self.window.set_caption(self.name)
            self.window.set_mouse_visible(False)

            try:
                # Icons. Currently hardcoded. Feel free to implement custom icons.
                path, filename = os.path.split(os.path.realpath(__file__))
                path = os.path.join(path, '..', '..', '..',
                                    'docs', 'source', 'static', 'qp-slm-notext-')
                img16x16 = pyglet.image.load(path + '16x16.png')
                img32x32 = pyglet.image.load(path + '32x32.png')
                self.window.set_icon(img16x16, img32x32)
            except Exception as e:
                print(e)

            # Set the viewpoint.
            proj = pyglet.window.Projection2D()
            proj.set(self.shape[1], self.shape[0], self.shape[1], self.shape[0])

            # Setup shapes
            texture_shape = tuple(np.power(2, np.ceil(np.log2(self.shape)))
                                .astype(np.int64))
            self.tex_shape_ratio = (float(self.shape[0])/float(texture_shape[0]),
                                    float(self.shape[1])/float(texture_shape[1]))
            B = 4

            # Setup buffers (texbuffer is power of 2 padded to init the memory in OpenGL)
            self.buffer = np.zeros(self.shape + (B,), dtype=np.uint8)
            N = int(self.shape[0] * self.shape[1] * B)
            self.cbuffer = (gl.GLubyte * N).from_buffer(self.buffer)

            texbuffer = np.zeros(texture_shape + (B,), dtype=np.uint8)
            Nt = int(texture_shape[0] * texture_shape[1] * B)
            texcbuffer = (gl.GLubyte * Nt).from_buffer(texbuffer)

            # Setup the texture
            gl.glEnable(gl.GL_TEXTURE_2D)
            self.texture = gl.GLuint()
            gl.glGenTextures(1, ctypes.byref(self.texture))
            gl.glBindTexture(gl.GL_TEXTURE_2D, self.texture.value)

            gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_GENERATE_MIPMAP, gl.GL_FALSE)
            gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_NEAREST)
            gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_NEAREST)

            # Malloc the OpenGL memory
            gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_RGBA8,
                            texture_shape[1], texture_shape[0],
                            0, gl.GL_BGRA, gl.GL_UNSIGNED_BYTE,
                            texcbuffer)

            # Make sure we can write to a subset of the memory (as we will do in the future)
            gl.glTexSubImage2D(gl.GL_TEXTURE_2D, 0, 0, 0,
                            self.shape[1], self.shape[0],
                            gl.GL_BGRA, gl.GL_UNSIGNED_BYTE,
                            self.cbuffer)

            # Cleanup
            gl.glBindTexture(gl.GL_TEXTURE_2D, 0)
            gl.glFlush()

            # Write nothing.
            self.write(phase=None)
        except:
            # If we failed, try to clean up before erroring.
            try:
                self.window.close()
            except:
                pass
            raise

        if verbose:
            print("success")

        # Warn the user if wav_um > wav_design_um
        if self.phase_scaling > 1:
            print(
                "Warning: Wavelength {} um is inaccessible to this SLM with "
                "design wavelength {} um".format(self.wav_um, self.wav_design_um)
            )

    def _write_hw(self, data):
        """Writes to screen. See :class:`.SLM`."""
        # Write to buffer (self.buffer is the same as self.cbuffer).
        # Unfortunately, OpenGL needs the data copied three times.
        np.copyto(self.buffer[:,:,0], data)
        np.copyto(self.buffer[:,:,1], data)
        np.copyto(self.buffer[:,:,2], data)

        # Setup texture variables.
        x1 = 0
        y1 = 0
        x2 = self.shape[1]
        y2 = self.shape[0]

        xa = 0
        ya = 0
        xb = self.tex_shape_ratio[1]
        yb = self.tex_shape_ratio[0]

        array = (gl.GLfloat * 32)(
            xa, ya, 0., 1.,         # tex coord,
            x1, y1, 0., 1.,         # real coord, ...
            xb, ya, 0., 1.,
            x2, y1, 0., 1.,
            xb, yb, 0., 1.,
            x2, y2, 0., 1.,
            xa, yb, 0., 1.,
            x1, y2, 0., 1.)

        # Update the texture.
        gl.glEnable(gl.GL_TEXTURE_2D)
        gl.glBindTexture(gl.GL_TEXTURE_2D, self.texture.value)
        gl.glTexSubImage2D(gl.GL_TEXTURE_2D, 0, 0, 0,
                        self.shape[1], self.shape[0],
                        gl.GL_RGBA, gl.GL_UNSIGNED_BYTE,
                        self.cbuffer)

        # Blit the texture.
        gl.glPushClientAttrib(gl.GL_CLIENT_VERTEX_ARRAY_BIT)
        gl.glInterleavedArrays(gl.GL_T4F_V4F, 0, array)
        gl.glDrawArrays(gl.GL_QUADS, 0, 4)
        gl.glPopClientAttrib()

        # Display the other side of the double buffer.
        # (with vsync enabled, this will block until the next frame is ready to display).
        self.window.flip()

    def close(self):
        """Closes frame. See :class:`.SLM`."""
        self.window.close()

    @staticmethod
    def info(verbose=True):
        """
        Get information about the available displays, their indexes, and their sizes.

        Parameters
        ----------
        verbose : bool
            Whether or not to print display information.

        Returns
        -------
        list of (int, (int, int, int, int)) tuples
            The number and geometry of each display.
        """
        # Note: in pyglet, the display is the full arrangement of screens,
        # unlike the terminology in other SLM subclasses
        display = pyglet.canvas.get_display()

        screens = display.get_screens()
        default = display.get_default_screen()
        windows = display.get_windows()

        def parse_screen(screen):
            return ("x={}, y={}, width={}, height={}"
                .format(screen.x, screen.y, screen.width, screen.height))
        def parse_screen_int(screen):
            return (screen.x, screen.y, screen.width, screen.height)
        def parse_window(window):
            x, y = window.get_location()
            return ("x={}, y={}, width={}, height={}"
                .format(x, y, window.width, window.height))

        default_str = parse_screen(default)

        window_strs = []
        for window in windows:
            window_strs.append(parse_window(window))

        if verbose:
            print('Display Positions:')
            print('#,  Position')

        screen_list = []

        for x, screen in enumerate(screens):
            screen_str = parse_screen(screen)

            main_bool = False
            window_bool = False

            if screen_str == default_str:
                main_bool = True
                screen_str += ' (main)'
            if screen_str in window_strs:
                window_bool = True
                screen_str += ' (has ScreenMirrored)'

            if verbose:
                print('{},  {}'.format(x, screen_str))

            screen_list.append((x, parse_screen_int(screen),
                                main_bool, window_bool))

        return screen_list



r"""
Common fit functions.
"""

import numpy as np
from scipy.special import factorial


# 1D

def linear(x, m, b):
    r"""
    For fitting a line.

    .. math:: y(x) = mx + b

    Parameters
    ----------
    x : numpy.ndarray
        Some independent variable.
    m : float
        Slope of the line.
    b : float
        :math:`y`-intercept of the line.

    Returns
    -------
    y : numpy.ndarray
        Line evaluated at all ``x``.
    """
    return m * x + b


def parabola(x, a, x0, y0):
    r"""
    For fitting a parabola.

    .. math:: y(x) = a(x - x_0)^2 + y_0

    Parameters
    ----------
    x : numpy.ndarray
        Some independent variable.
    a : float
        Strength of the parabola.
    x0 : float
        :math:`x` position of the vertex.
    y0 : float
        :math:`y` position of the vertex.

    Returns
    -------
    y : numpy.ndarray
        Line evaluated at all ``x``.
    """
    return a * np.square(x - x0) + y0


def hyperbola(z, w0, z0, zr):
    r"""
    For fitting a hyperbola.

    .. math:: w(z) = w_0 \sqrt{1 + \left[\frac{z - z_0}{z_R}\right]^2}.

    Parameters
    ----------
    z : numpy.ndarray
        Distance.
    w0 : float
        Beamradius at :math:`z = z_0`.
    z0 : float
        Plane of focus :math:`x_0`, the center of the hyperbola.
    zr : float
        Rayleigh length :math:`z_R`, the depth of focus.

    Returns
    -------
    w : numpy.ndarray
        Hyperbola evaluated at all ``z``.
    """
    return w0 * np.sqrt(1 + np.square((z - z0) / zr))


def cos(x, b, a, c, k=1):
    r"""
    For fitting an offset sinusoid.

    .. math:: y(x) = c + \frac{a}{2} \left[1+\cos(kx-b) \right].

    Parameters
    ----------
    x : numpy.ndarray
        Phase in radians.
    b : float
        Phase offset.
    a : float
        Peak amplitude.
    c : float
        Amplitude offset.
    k : float
        Phase scale factor. Default is 1.

    Returns
    -------
    y : numpy.ndarray
        Cosine fit evaluated at all ``x``.
    """
    return a * 0.5 * (1 + np.cos(k * x - b)) + c


def lorentzian(x, x0, a, c, Q):
    r"""
    For fitting a resonance.
    There are many ways to describe a Lorentzian. Commonly, a full-width-half-maximum
    definition is used. Here, with roots in photonic crystal cavities, we use a
    form defined using the quality factor :math:`Q` of the resonance.

    .. math:: y(x) = c + \frac{a}{1 + \left[\frac{x - x_0}{x_0/2Q}\right]^2}.

    Parameters
    ----------
    x : numpy.ndarray
        Points to fit upon.
    x0 : float
        Center wavelength.
    a : float
        Amplitude.
    c : float
        Constant offset.
    Q : float
        Quality factor.

    Returns
    -------
    y : numpy.ndarray
        Lorentzian fit evaluated at all ``x``.
    """
    return a / (1 + ((x - x0) / (x0 / Q / 2)) ** 2) + c


def lorentzian_jacobian(x, x0, a, c, Q):
    """
    Jacobian of :meth:`lorentzian`.

    Parameters
    ----------
    x : numpy.ndarray
        Wavelength.
    x0 : float
        Center wavelength.
    a : float
        Amplitude.
    c : float
        constant offset.
    Q : float
        Quality factor.

    Returns
    -------
    gradf : numpy.ndarray
        Jacobian of Lorentzian fit evaluated at all ``x``.
    """
    return np.array(
        [
            a
            * 8
            * Q ** 2
            * (x - x0)
            / x0 ** 2
            * (1 + (x - x0) / x0)
            / (1 + 4 * (Q * (x - x0) / x0) ** 2) ** 2,
            1 / (1 + 4 * (Q * (x - x0) / x0) ** 2),
            -8
            * Q
            * a
            * ((x - x0) / x0) ** 2
            / (1 + 4 * (Q * (x - x0) / x0) ** 2) ** 2,
            1 - 1 / (1 + 4 * (Q * (x - x0) / x0) ** 2),
        ]
    ).T


def gaussian(x, x0, a, c, w):
    r"""
    For fitting a 1D Gaussian.

    .. math:: y(x) = c + a \exp \left[-\frac{(x-x_0)^2}{2w^2}\right].

    Parameters
    ----------
    x : numpy.ndarray
        Points to fit upon.
    x0 : float
        Positional offset.
    a : float
        Amplitude.
    c : float
        constant offset.
    w : float
        The standard deviation of the normal distribution.
        Equivalent to the :math:`1/e` radius.
        This is related to the full width at half maximum (FWHM)
        by a factor of :math:`2\sqrt{2\ln{2}}`.

    Returns
    -------
    y : numpy.ndarray
        Gaussian fit evaluated at all ``x``.
    """
    return c + a * np.exp(-.5 * np.square((x - x0) * (1/w)))


# 2D

def gaussian2d(xy, x0, y0, a, c, wx, wy, wxy=0):
    r"""
    For fitting a 2D Gaussian.
    There are two cases which depend on the shear variance ``wxy``
    (equivalent to :math:`M_{11}`;
    see :meth:`~slmsuite.holography.analysis.image_moment()`).

    When ``wxy`` is zero, then we have the usual form of a Gaussian:

    .. math:: z(x,y) = c + a \exp \left[
                                -\frac{(x-x_0)^2}{2w_x^2} -
                                \frac{(y-y_0)^2}{2w_y^2}
                                \right].

    When ``wxy`` is nonzero, the 2D Gaussian to have second
    order central moments (equivalent to variance;
    see :meth:`~slmsuite.holography.analysis.image_variances()`) satisfying:

    .. math::   M =
                \begin{bmatrix}
                    M_{20} & M_{11} \\
                    M_{11} & M_{02}
                \end{bmatrix}
                =
                \begin{bmatrix}
                    w_x^2 & w_{xy} \\
                    w_{xy} & w_y^2
                \end{bmatrix}.

    The following satisfies satisfying the above condition:

    .. math:: z(x,y) = c + a \exp \left[
                                -\frac{1}{2}\left(
                                K_{00}(x-x_0)^2 +
                                2*K_{10}(x-x_0)(y-y_0) +
                                K_{11}(y-y_0)^2
                                \right)
                                \right].

    Where

    .. math:: K =
                \begin{bmatrix}
                    K_{00} & K_{10} \\
                    K_{10} & K_{11}
                \end{bmatrix}
                = M^{-1}.

    Caution
    ~~~~~~~
    The shear variance :math:`w_{xy}` does not have the same units as the widths
    :math:`w_x` and :math:`w_y`. The widths have units of space, whereas the shear
    variance has units of squared space.

    Note
    ~~~~
    The shear variance ``wxy`` is currently bounded to magnitudes below ``wx*wy``.
    Higher values lead to solutions which cannot be normalized.
    When ``wxy = wx*wy``, this distribution is a line (an ellipse with zeroed minor
    axis).

    Parameters
    ----------
    xy : numpy.ndarray
        Points to fit upon. ``(x, y)`` form.
    x0, y0 : float
        Vector offset.
    a : float
        Amplitude.
    c : float
        constant offset.
    wx, wy : float
        The standard deviation of the normal distribution.
        Equivalent to the :math:`1/e` radius.
        This is related to the full width at half maximum (FWHM)
        by a factor of :math:`2\sqrt{2\ln{2}}`.
    wxy : float
        Shear variance. See above.

    Returns
    -------
    z : numpy.ndarray
        Gaussian fit evaluated at all ``(x,y)`` in ``xy``.
    """
    x = xy[0] - x0
    y = xy[1] - y0

    wxy = np.sign(wxy) * np.min([np.abs(wxy), wx*wy])

    try:
        M = np.linalg.inv([[wx*wx, wxy], [wxy, wy*wy]])
    except np.linalg.LinAlgError:
        M = np.array([[1/wx/wx, 0], [0, 1/wy/wy]])

    argument = np.square(x) * M[0,0] + np.square(y) * M[1,1] + 2 * x * y * M[1,0]

    return c + a * np.exp(-.5 * argument)


def tophat2d(xy, x0, y0, R, a=1, c=0):
    r"""
    For fitting a 2D tophat distribution.

    .. math:: z(x,y) =  \left\{
                            \begin{array}{ll}
                                a + c, & x^2 + y^2 < R^2 \\
                                c, & \text{ otherwise}.
                            \end{array}
                        \right.

    Parameters
    ----------
    xy : numpy.ndarray
        Points to fit upon (x, y).
    x0, y0 : float
        Vector offset.
    R : float
        Active radius of the tophat.
    a : float
        Amplitude.
    c : float
        Offset.

    Returns
    -------
    z : numpy.ndarray
        Tophat fit evaluated at all ``(x,y)`` in ``xy``.
    """
    x = xy[0] - x0
    y = xy[1] - y0
    return np.where(np.square(x) + np.square(y) <= R*R, a+c, c)


def sinc2d(xy, x0, y0, R, a=1, b=0, c=0, d=0, kx=0, ky=0):
    r"""
    For fitting a 2D rectangular :math:`\text{sinc}^2` distribution, potentially with a sinusoidal modulation.

    .. math:: z(x,y) =  d + \left(c + \frac{a}{2} \left[1+\cos(k_xx+k_yy-b) \right]\right) *
                        \text{sinc}^2(\pi (x-x_0) / R) * \text{sinc}^2(\pi (y-y_0) / R).

    where

    .. math:: \text{sinc}(x) = \frac{\sin(x)}{x}

    Parameters
    ----------
    xy : numpy.ndarray
        Points to fit upon (x, y).
    x0, y0 : float
        Vector offset.
    R : float
        Square radius of the sinc (radius of the first zero).
    a : float
        Peak amplitude.
    b : float
        Phase offset.
    c : float
        Sinusoidal amplitude offset.
    d : float
        Global offset.
    kx, ky : float
        Vector phase scale factor. Default is 0.

    Returns
    -------
    z : numpy.ndarray
        Rectangular sinc fit evaluated at all ``(x,y)`` in ``xy``.
    """
    x = xy[0] - x0
    y = xy[1] - y0

    return np.square(np.sinc((1 / R) * x) * np.sinc((1 / R) * y)) \
            * (a * 0.5 * (1 + np.cos(kx * x + ky * y - b)) + c) + d


# sinc variations

def _sinc2d_nomod(xy, x0, y0, R, a=1, d=0):
    r"""
    For fitting a 2D rectangular sinc distribution, potentially with a sinusoidal modulation.

    .. math:: z(x,y) =  d + a * \text{sinc}^2(\pi (x-x_0) / R) * \text{sinc}^2(\pi (y-y_0) / R).

    where

    .. math:: \text{sinc}(x) = \frac{\sin(x)}{x}

    Parameters
    ----------
    xy : numpy.ndarray
        Points to fit upon (x, y).
    x0, y0 : float
        Vector offset.
    R : float
        Square radius of the sinc (radius of the first zero).
    a : float
        Peak amplitude.
    d : float
        Global offset.

    Returns
    -------
    z : numpy.ndarray
        Rectangular sinc fit evaluated at all ``(x,y)`` in ``xy``.
    """
    return (
        a * np.square(np.sinc((1 / R) * (xy[0] - x0)) * np.sinc((1 / R) * (xy[1] - y0))) + d
    )

def _sinc2d_nomod_taylor(xy, x0, y0, R, a=1, d=0):
    r"""
    For fitting a 2D rectangular sinc distribution, potentially with a sinusoidal modulation.

    .. math:: z(x,y) =  d + a * \text{sinc}^2(\pi (x-x_0) / R) * \text{sinc}^2(\pi (y-y_0) / R).

    where

    .. math:: \text{sinc}(x) = \frac{\sin(x)}{x}

    Parameters
    ----------
    xy : numpy.ndarray
        Points to fit upon (x, y).
    x0, y0 : float
        Vector offset.
    R : float
        Square radius of the sinc (radius of the first zero).
    a : float
        Peak amplitude.
    d : float
        Global offset.

    Returns
    -------
    z : numpy.ndarray
        Rectangular sinc fit evaluated at all ``(x,y)`` in ``xy``.
    """
    return (
        a * np.square(_sinc_taylor((1 / R) * (xy[0] - x0)) * _sinc_taylor((1 / R) * (xy[1] - y0))) + d
    )

def _sinc2d_centered(xy, R, a=1, b=0, c=0, d=0, kx=0, ky=0):
    r"""
    For fitting a 2D rectangular sinc distribution, potentially with a sinusoidal modulation.

    .. math:: z(x,y) =  d + \left(c + \frac{a}{2} \left[1+\cos(k_xx+k_yy-b) \right]\right) *
                        \text{sinc}^2(\pi (x-x_0) / R) * \text{sinc}^2(\pi (y-y_0) / R).

    where

    .. math:: \text{sinc}(x) = \frac{\sin(x)}{x}

    Parameters
    ----------
    xy : numpy.ndarray
        Points to fit upon (x, y).
    R : float
        Square radius of the sinc (radius of the first zero).
    a : float
        Peak amplitude.
    b : float
        Phase offset.
    c : float
        Sinusoidal amplitude offset.
    d : float
        Global offset.
    kx, ky : float
        Vector phase scale factor. Default is 1.

    Returns
    -------
    z : numpy.ndarray
        Rectangular sinc fit evaluated at all ``(x,y)`` in ``xy``.
    """
    return (
        np.square(np.sinc((1 / R) * xy[0]) * np.sinc((1 / R) * xy[1]))
        * (a * 0.5 * (1 + np.cos(kx * xy[0] + ky * xy[1] - b)) + c) + d
    )

def _sinc2d_centered_taylor(xy, R, a=1, b=0, c=0, d=0, kx=0, ky=0):
    r"""
    For fitting a 2D rectangular sinc distribution, potentially with a sinusoidal modulation.

    .. math:: z(x,y) =  d + \left(c + \frac{a}{2} \left[1+\cos(k_xx+k_yy-b) \right]\right) *
                        \text{sinc}^2(\pi (x-x_0) / R) * \text{sinc}^2(\pi (y-y_0) / R).

    where

    .. math:: \text{sinc}(x) = \frac{\sin(x)}{x}

    Parameters
    ----------
    xy : numpy.ndarray
        Points to fit upon (x, y).
    R : float
        Square radius of the sinc (radius of the first zero).
    a : float
        Peak amplitude.
    b : float
        Phase offset.
    c : float
        Sinusoidal amplitude offset.
    d : float
        Global offset.
    kx, ky : float
        Vector phase scale factor. Default is 1.

    Returns
    -------
    z : numpy.ndarray
        Rectangular sinc fit evaluated at all ``(x,y)`` in ``xy``.
    """
    return (
        np.square(_sinc_taylor((1 / R) * xy[0]) * _sinc_taylor((1 / R) * xy[1]))
        * (a * 0.5 * (1 + np.cos(kx * xy[0] + ky * xy[1] - b)) + c) + d
    )

def _sinc_taylor(x, order=12):
    """
    Taylor series approximation for sinc. We use the numpy normalization.

    Parameters
    ----------
    x : numpy.ndarray
        Array to approximate sinc(x) upon.
    order : int
        Order of 12 approximates well up to the second zero.
    """
    squared = np.square(np.pi * x)
    monomial = squared.copy()
    result = 1

    for n in range(2, order+2, 2):
        if n != 2:
            monomial *= squared
        result += monomial * ((-1 if n % 4 == 2 else 1) / factorial(n+1))

    return result

def _sinc2d_centered_jacobian(xy, R, a=1, b=0, c=0, d=0, kx=0, ky=0):
    r"""
    Jacobian of :meth:`.sinc2d_centered()`.

    Returns
    -------
    z : numpy.ndarray
        Rectangular sinc fit evaluated at all ``(x,y)`` in ``xy``.
    """
    scx = np.sinc((1 / R) * xy[0])
    scy = np.sinc((1 / R) * xy[1])
    cx = np.cos((1 / R) * xy[0])
    cy = np.cos((1 / R) * xy[1])
    sinc_term = np.square(scx * scy)
    cos_term = 0.5 * (1 + np.cos(kx * xy[0] + ky * xy[1] - b))
    dcos_term = -0.5 * np.sin(kx * xy[0] + ky * xy[1] - b)
    return np.vstack((
        # R
        (2 / R) * scx * scy * (scx * (scy - cy) + scy * (scx - cx))
        * (a * cos_term + c),
        # a
        sinc_term * cos_term,
        # b
        -sinc_term * a * dcos_term,
        # c
        sinc_term,
        # d
        np.full_like(xy[0], 1),
        # kx
        xy[0] * sinc_term * a * dcos_term,
        # ky
        xy[1] * sinc_term * a * dcos_term,
    )).T




"""
A slightly-modified Santec header file for Santec SLMs.
Modified for dll handling for other versions of python,
along with error handling and status dictionaries.
See santec.py.
"""
import os
from ctypes import *
STRING = c_char_p
from ctypes.wintypes import DWORD
from ctypes.wintypes import ULONG
from ctypes.wintypes import WORD
from ctypes.wintypes import BYTE
from ctypes.wintypes import BOOL
from ctypes.wintypes import BOOLEAN
from ctypes.wintypes import LPCSTR
from ctypes.wintypes import HANDLE
from ctypes.wintypes import LONG
from ctypes.wintypes import UINT
from ctypes.wintypes import LPSTR
from ctypes.wintypes import LPCSTR
from ctypes.wintypes import LPCWSTR
from ctypes.wintypes import FILETIME

if hasattr(os, "add_dll_directory"):    # python >= 3.8
    os.add_dll_directory(os.getcwd())
    os.add_dll_directory(os.path.dirname(os.path.abspath(__file__)))
    _libname = "SLMFunc.dll"
    _libraries = {}
    _libraries[_libname] = WinDLL(_libname)
else:                                   # python < 3.8
    _libname = "SLMFunc.dll"
    _libpath = os.path.dirname(os.path.abspath(__file__))
    os.environ['PATH'] = _libpath + os.pathsep + os.environ['PATH']
    _libraries = {}
    _libraries[_libname] = windll.LoadLibrary(_libname)

# ctypes
USHORT = c_ushort
SHORT = c_short
UCHAR = c_ubyte
WCHAR = c_wchar
LPBYTE = POINTER(c_ubyte)
LPUSHORT = POINTER(c_ushort)
CHAR = c_char
LPBOOL = POINTER(c_int)
PUCHAR = POINTER(c_ubyte)
PCHAR = STRING
PVOID = c_void_p
INT = c_int
LPTSTR = STRING
LPDWORD = POINTER(c_uint32)
LPWORD = POINTER(WORD)
PULONG = POINTER(ULONG)
LPVOID = PVOID
VOID = None
ULONGLONG = c_ulonglong
SLM_STATUS = c_int32

# SLM_Ctrl_ReadEDO coding
SLM_DRIVEBOARD_ERROR = {
    0x01 : "Startup error 1 (Drive board)",
    0x02 : "Startup error 2 (Drive board)",
    0x04 : "Video signal error (No signal)",
    0x08 : "Drive board temperature error (70°C or higher)"
}
SLM_OPTIONBOARD_ERROR = {
    0x01 : "Startup error 1 (Option board)",
    0x02 : "Startup error 2 (Option board)",
    0x04 : "Voltage level error (DC 5.0V)",
    0x08 : "Option board temperature error (70°C or higher)"
}

# SLM_STATUS
# From Table 3.4-1 SLM_STATUS. Some notes are modified to include implied meaning.
# { value : ("name", "note") }
SLM_STATUS_DICT = {
    0 :      ("SLM_OK", "All good!"),
    1 :      ("SLM_NG", "NG"),
    2 :      ("SLM_BS", "SLM is busy."),
    3 :      ("SLM_ER", "Parameter error."),
    -1 :     ("SLM_INVAID_MONITOR", "Could not find specified display number."),
    -2 :     ("SLM_NOT_OPEN_MONITOR", "Display has not been opened."),
    -3 :     ("SLM_OPEN_WINDOW_ERR", "Window open error."),
    -4 :     ("SLM_DATA_FORMAT_ERR", "Data format error."),
    -101 :   ("SLM_FILE_READ_ERR", "File contained data over 1023."),
    -200 :   ("SLM_NOT_OPEN_USB", "USB is not open."),
    -1000 :  ("SLM_OTHER_ERROR", "Other error."),
    -10001 : ("FT_INVALID_HANDLE", "USB driver error."),
    -10002 : ("FT_DEVICE_NOT_FOUND", "Device not found. Check device's power. If connected, reset the power."),
    -10003 : ("FT_DEVICE_NOT_OPENED", "Already opened."),
    -10004 : ("FT_IO_ERROR", "USB driver error."),
    -10005 : ("FT_INSUFFICIENT_RESOURCES", "USB driver error."),
    -10006 : ("FT_INVALID_PARAMETER", "USB driver error."),
    -10007 : ("FT_INVALID_BAUD_RATE", "USB driver error."),
    -10008 : ("FT_DEVICE_NOT_OPENED_FOR_ERASE", "USB driver error."),
    -10009 : ("FT_DEVICE_NOT_OPENED_FOR_WRITE", "USB driver error."),
    -10010 : ("FT_FAILED_TO_WRITE_DEVICE",      "USB driver error."),
    -10011 : ("FT_EEPROM_READ_FAILED", "USB driver error."),
    -10012 : ("FT_EEPROM_WRITE_FAILED", "USB driver error."),
    -10013 : ("FT_EEPROM_ERASE_FAILED", "USB driver error."),
    -10014 : ("FT_EEPROM_NOT_PRESENT", "USB driver error."),
    -10015 : ("FT_EEPROM_NOT_PROGRAMMED", "USB driver error."),
    -10016 : ("FT_INVALID_ARGS", "USB driver error."),
    -10017 : ("FT_NOT_SUPPORTED", "USB driver error."),
    -10018 : ("FT_NO_MORE_ITEMS", "USB driver error."),
    -10019 : ("FT_TIMEOUT", "USB driver error."),
    -10020 : ("FT_OPERATION_ABORTED", "USB driver error."),
    -10021 : ("FT_RESERVED_PIPE", "USB driver error."),
    -10022 : ("FT_INVALID_CONTROL_REQUEST_DIRECTION", "USB driver error."),
    -10023 : ("FT_INVALID_CONTROL_REQUEST_TYPE", "USB driver error."),
    -10024 : ("FT_IO_PENDING", "USB driver error."),
    -10025 : ("FT_IO_INCOMPLETE", "USB driver error."),
    -10026 : ("FT_HANDLE_EOF", "USB driver error."),
    -10027 : ("FT_BUSY", "USB driver error."),
    -10028 : ("FT_NO_SYSTEM_RESOURCES", "USB driver error."),
    -10029 : ("FT_DEVICE_LIST_NOT_READY", "USB driver error."),
    -10030 : ("FT_DEVICE_NOT_CONNECTED", "USB driver error."),
    -10031 : ("FT_INCORRECT_DEVICE_PATH", "USB driver error."),
    -10032 : ("FT_OTHER_ERROR", "USB driver error.")
}

SLM_Disp_Info = _libraries[_libname].SLM_Disp_Info
SLM_Disp_Info.restype = SLM_STATUS
SLM_Disp_Info.argtypes = [DWORD, LPUSHORT, LPUSHORT]
SLM_Disp_Info.__doc__ = (
    """SLM_Disp_Info(DWORD DisplayNumber, USHORT *width, USHORT *height)"""
)

SLM_Disp_Info2 = _libraries[_libname].SLM_Disp_Info2
SLM_Disp_Info2.restype = SLM_STATUS
SLM_Disp_Info2.argtypes = [DWORD, LPUSHORT, LPUSHORT, LPSTR]
SLM_Disp_Info2.__doc__ = """SLM_Disp_Info2(DWORD DisplayNumber, USHORT *width, USHORT *height, LPSTR DisplayName )"""

SLM_Disp_Open = _libraries[_libname].SLM_Disp_Open
SLM_Disp_Open.restype = SLM_STATUS
SLM_Disp_Open.argtypes = [DWORD]
SLM_Disp_Open.__doc__ = """SLM_Disp_Open(DWORD DisplayNumber)"""

SLM_Disp_Close = _libraries[_libname].SLM_Disp_Close
SLM_Disp_Close.restype = SLM_STATUS
SLM_Disp_Close.argtypes = [DWORD]
SLM_Disp_Close.__doc__ = """SLM_Disp_Close(DWORD DisplayNumber)"""

SLM_Disp_GrayScale = _libraries[_libname].SLM_Disp_GrayScale
SLM_Disp_GrayScale.restype = SLM_STATUS
SLM_Disp_GrayScale.argtypes = [DWORD, DWORD, USHORT]
SLM_Disp_GrayScale.__doc__ = (
    """SLM_Disp_GrayScale(DWORD DisplayNumber, DWORD Flags, USHORT GrayScale)"""
)

# SLM_Disp_BMP = _libraries[_libname].SLM_Disp_BMP
# SLM_Disp_BMP.restype = SLM_STATUS
# SLM_Disp_BMP.argtypes = [DWORD, DWORD, HBITMAP]
# SLM_Disp_BMP.__doc__ = \
# """SLM_Disp_BMP(DWORD DisplayNumber, DWORD Flags, HBITMAP bmp)"""

SLM_Disp_Data = _libraries[_libname].SLM_Disp_Data
SLM_Disp_Data.restype = SLM_STATUS
SLM_Disp_Data.argtypes = [DWORD, USHORT, USHORT, DWORD, c_void_p]
SLM_Disp_Data.__doc__ = """SLM_Disp_Data(DWORD DisplayNumber, USHORT width, USHORT height, DWORD Flags, USHORT* data)"""

SLM_Disp_ReadBMP = _libraries[_libname].SLM_Disp_ReadBMP
SLM_Disp_ReadBMP.restype = SLM_STATUS
SLM_Disp_ReadBMP.argtypes = [DWORD, DWORD, LPCWSTR]
SLM_Disp_ReadBMP.__doc__ = (
    """SLM_Disp_ReadBMP(DWORD DisplayNumber, DWORD Flags, LPCWSTR FileName)"""
)

SLM_Disp_ReadCSV = _libraries[_libname].SLM_Disp_ReadCSV
SLM_Disp_ReadCSV.restype = SLM_STATUS
SLM_Disp_ReadCSV.argtypes = [DWORD, DWORD, LPCWSTR]
SLM_Disp_ReadCSV.__doc__ = (
    """SLM_Disp_ReadCSV(DWORD DisplayNumber, DWORD Flags, LPCWSTR FileName)"""
)

SLM_Disp_ReadBMP_A = _libraries[_libname].SLM_Disp_ReadBMP_A
SLM_Disp_ReadBMP_A.restype = SLM_STATUS
SLM_Disp_ReadBMP_A.argtypes = [DWORD, DWORD, LPCSTR]
SLM_Disp_ReadBMP_A.__doc__ = (
    """SLM_Disp_ReadBMP_A(DWORD DisplayNumber, DWORD Flags, LPCSTR FileName)"""
)

SLM_Disp_ReadCSV_A = _libraries[_libname].SLM_Disp_ReadCSV_A
SLM_Disp_ReadCSV_A.restype = SLM_STATUS
SLM_Disp_ReadCSV_A.argtypes = [DWORD, DWORD, LPCSTR]
SLM_Disp_ReadCSV_A.__doc__ = (
    """SLM_Disp_ReadCSV_A(DWORD DisplayNumber, DWORD Flags, LPCSTR FileName)"""
)


SLM_Ctrl_Open = _libraries[_libname].SLM_Ctrl_Open
SLM_Ctrl_Open.restype = SLM_STATUS
SLM_Ctrl_Open.argtypes = [DWORD]
SLM_Ctrl_Open.__doc__ = """SLM_Ctrl_Open(DWORD SLMNumber)"""

SLM_Ctrl_Close = _libraries[_libname].SLM_Ctrl_Close
SLM_Ctrl_Close.restype = SLM_STATUS
SLM_Ctrl_Close.argtypes = [DWORD]
SLM_Ctrl_Close.__doc__ = """SLM_Ctrl_Close(DWORD SLMNumber)"""

SLM_Ctrl_Read = _libraries[_libname].SLM_Ctrl_Read
SLM_Ctrl_Read.restype = SLM_STATUS
SLM_Ctrl_Read.argtypes = [DWORD, LPBYTE, LPUSHORT]
SLM_Ctrl_Read.__doc__ = (
    """SLM_Ctrl_Read(DWORD SLMNumber, BYTE* recv, USHORT* recv_len)"""
)

SLM_Ctrl_WriteVI = _libraries[_libname].SLM_Ctrl_WriteVI
SLM_Ctrl_WriteVI.restype = SLM_STATUS
SLM_Ctrl_WriteVI.argtypes = [DWORD, DWORD]
SLM_Ctrl_WriteVI.__doc__ = """SLM_Ctrl_WriteVI(DWORD SLMNumber, DWORD mode)"""

SLM_Ctrl_ReadVI = _libraries[_libname].SLM_Ctrl_ReadVI
SLM_Ctrl_ReadVI.restype = SLM_STATUS
SLM_Ctrl_ReadVI.argtypes = [DWORD, LPDWORD]
SLM_Ctrl_ReadVI.__doc__ = """SLM_Ctrl_ReadVI(DWORD SLMNumber, DWORD *mode)"""

SLM_Ctrl_WriteWL = _libraries[_libname].SLM_Ctrl_WriteWL
SLM_Ctrl_WriteWL.restype = SLM_STATUS
SLM_Ctrl_WriteWL.argtypes = [DWORD, DWORD, DWORD]
SLM_Ctrl_WriteWL.__doc__ = (
    """SLM_Ctrl_WriteWL(DWORD SLMNumber, DWORD wavelength, float phase)"""
)

SLM_Ctrl_ReadWL = _libraries[_libname].SLM_Ctrl_ReadWL
SLM_Ctrl_ReadWL.restype = SLM_STATUS
SLM_Ctrl_ReadWL.argtypes = [DWORD, LPDWORD, LPDWORD]
SLM_Ctrl_ReadWL.__doc__ = (
    """SLM_Ctrl_ReadWL(DWORD SLMNumber, DWORD *wavelength, float *phase)"""
)

SLM_Ctrl_WriteAW = _libraries[_libname].SLM_Ctrl_WriteAW
SLM_Ctrl_WriteAW.restype = SLM_STATUS
SLM_Ctrl_WriteAW.argtypes = [DWORD]
SLM_Ctrl_WriteAW.__doc__ = """SLM_Ctrl_WriteAW(DWORD SLMNumber)"""

SLM_Ctrl_WriteTI = _libraries[_libname].SLM_Ctrl_WriteTI
SLM_Ctrl_WriteTI.restype = SLM_STATUS
SLM_Ctrl_WriteTI.argtypes = [DWORD, DWORD]
SLM_Ctrl_WriteTI.__doc__ = """SLM_Ctrl_WriteTI(DWORD SLMNumber, DWORD onoff)"""

SLM_Ctrl_ReadTI = _libraries[_libname].SLM_Ctrl_ReadTI
SLM_Ctrl_ReadTI.restype = SLM_STATUS
SLM_Ctrl_ReadTI.argtypes = [DWORD, LPDWORD]
SLM_Ctrl_ReadTI.__doc__ = """SLM_Ctrl_ReadTI(DWORD SLMNumber, DWORD *onoff)"""

SLM_Ctrl_WriteTM = _libraries[_libname].SLM_Ctrl_WriteTM
SLM_Ctrl_WriteTM.restype = SLM_STATUS
SLM_Ctrl_WriteTM.argtypes = [DWORD, DWORD]
SLM_Ctrl_WriteTM.__doc__ = """SLM_Ctrl_WriteTM(DWORD SLMNumber, DWORD onoff)"""

SLM_Ctrl_ReadTM = _libraries[_libname].SLM_Ctrl_ReadTM
SLM_Ctrl_ReadTM.restype = SLM_STATUS
SLM_Ctrl_ReadTM.argtypes = [DWORD, LPDWORD]
SLM_Ctrl_ReadTM.__doc__ = """SLM_Ctrl_ReadTM(DWORD SLMNumber, DWORD *onoff)"""

SLM_Ctrl_ReadTM = _libraries[_libname].SLM_Ctrl_ReadTM
SLM_Ctrl_ReadTM.restype = SLM_STATUS
SLM_Ctrl_ReadTM.argtypes = [DWORD, DWORD]
SLM_Ctrl_ReadTM.__doc__ = """SLM_Ctrl_WriteTC(DWORD SLMNumber, DWORD order)"""

SLM_Ctrl_ReadTC = _libraries[_libname].SLM_Ctrl_ReadTC
SLM_Ctrl_ReadTC.restype = SLM_STATUS
SLM_Ctrl_ReadTC.argtypes = [DWORD, LPDWORD]
SLM_Ctrl_ReadTC.__doc__ = """SLM_Ctrl_ReadTC(DWORD SLMNumber, DWORD *order)"""

SLM_Ctrl_WriteTS = _libraries[_libname].SLM_Ctrl_WriteTS
SLM_Ctrl_WriteTS.restype = SLM_STATUS
SLM_Ctrl_WriteTS.argtypes = [DWORD]
SLM_Ctrl_WriteTS.__doc__ = """SLM_Ctrl_WriteTS(DWORD SLMNumber)"""

SLM_Ctrl_WriteMC = _libraries[_libname].SLM_Ctrl_WriteMC
SLM_Ctrl_WriteMC.restype = SLM_STATUS
SLM_Ctrl_WriteMC.argtypes = [DWORD, DWORD]
SLM_Ctrl_WriteMC.__doc__ = """SLM_Ctrl_WriteMC(DWORD SLMNumber, DWORD MemoryNumber)"""

SLM_Ctrl_WriteMI = _libraries[_libname].SLM_Ctrl_WriteMI
SLM_Ctrl_WriteMI.restype = SLM_STATUS
SLM_Ctrl_WriteMI.argtypes = [DWORD, DWORD, USHORT, USHORT, DWORD, LPUSHORT]
SLM_Ctrl_WriteMI.__doc__ = """SLM_Ctrl_WriteMI(DWORD SLMNumber, DWORD MemoryNumber, USHORT width, USHORT height, DWORD Flags, USHORT* data)"""

SLM_Ctrl_WriteMI_BMP = _libraries[_libname].SLM_Ctrl_WriteMI_BMP
SLM_Ctrl_WriteMI_BMP.restype = SLM_STATUS
SLM_Ctrl_WriteMI_BMP.argtypes = [DWORD, DWORD, DWORD, LPCWSTR]
SLM_Ctrl_WriteMI_BMP.__doc__ = """SLM_Ctrl_WriteMI_BMP(DWORD SLMNumber, DWORD MemoryNumber, DWORD Flags, LPCWSTR FileName)"""

SLM_Ctrl_WriteMI_CSV = _libraries[_libname].SLM_Ctrl_WriteMI_CSV
SLM_Ctrl_WriteMI_CSV.restype = SLM_STATUS
SLM_Ctrl_WriteMI_CSV.argtypes = [DWORD, DWORD, DWORD, LPCWSTR]
SLM_Ctrl_WriteMI_CSV.__doc__ = """SLM_Ctrl_WriteMI_CSV(DWORD SLMNumber, DWORD MemoryNumber, DWORD Flags, LPCWSTR FileName)"""

# SLM_Ctrl_WriteMI_BMP_A = _libraries[_libname].SLM_Ctrl_WriteMI_BMP_A
# SLM_Ctrl_WriteMI_BMP_A.restype = SLM_STATUS
# SLM_Ctrl_WriteMI_BMP_A.argtypes = [DWORD, DWORD, DWORD, LPCSTR]
# SLM_Ctrl_WriteMI_BMP_A.__doc__ = \
# """SLM_Ctrl_WriteMI_BMP_A(DWORD SLMNumber, DWORD MemoryNumber, DWORD Flags, LPCSTR FileName)"""

SLM_Ctrl_WriteMI_CSV_A = _libraries[_libname].SLM_Ctrl_WriteMI_CSV_A
SLM_Ctrl_WriteMI_CSV_A.restype = SLM_STATUS
SLM_Ctrl_WriteMI_CSV_A.argtypes = [DWORD, DWORD, DWORD, LPCSTR]
SLM_Ctrl_WriteMI_CSV_A.__doc__ = """SLM_Ctrl_WriteMI_CSV_A(DWORD SLMNumber, DWORD MemoryNumber, DWORD Flags, LPCSTR FileName)"""

SLM_Ctrl_WriteME = _libraries[_libname].SLM_Ctrl_WriteME
SLM_Ctrl_WriteME.restype = SLM_STATUS
SLM_Ctrl_WriteME.argtypes = [DWORD, DWORD]
SLM_Ctrl_WriteME.__doc__ = """SLM_Ctrl_WriteME(DWORD SLMNumber, DWORD MemoryNumber)"""

SLM_Ctrl_WriteMT = _libraries[_libname].SLM_Ctrl_WriteMT
SLM_Ctrl_WriteMT.restype = SLM_STATUS
SLM_Ctrl_WriteMT.argtypes = [DWORD, DWORD, DWORD]
SLM_Ctrl_WriteMT.__doc__ = (
    """SLM_Ctrl_WriteMT(DWORD SLMNumber, DWORD TableNumber, DWORD MemoryNumber)"""
)

SLM_Ctrl_ReadMS = _libraries[_libname].SLM_Ctrl_ReadMS
SLM_Ctrl_ReadMS.restype = SLM_STATUS
SLM_Ctrl_ReadMS.argtypes = [DWORD, DWORD, c_void_p]
SLM_Ctrl_ReadMS.__doc__ = (
    """SLM_Ctrl_ReadMS(DWORD SLMNumber, DWORD TableNumber, DWORD *MemoryNumber)"""
)

SLM_Ctrl_WriteMR = _libraries[_libname].SLM_Ctrl_WriteMR
SLM_Ctrl_WriteMR.restype = SLM_STATUS
SLM_Ctrl_WriteMR.argtypes = [DWORD, DWORD, DWORD]
SLM_Ctrl_WriteMR.__doc__ = (
    """SLM_Ctrl_WriteMR(DWORD SLMNumber, DWORD TableNumber1, DWORD TableNumber2)"""
)

SLM_Ctrl_ReadMR = _libraries[_libname].SLM_Ctrl_ReadMR
SLM_Ctrl_ReadMR.restype = SLM_STATUS
SLM_Ctrl_ReadMR.argtypes = [DWORD, c_void_p, c_void_p]
SLM_Ctrl_ReadMR.__doc__ = (
    """SLM_Ctrl_ReadMR(DWORD SLMNumber, DWORD *TableNumber1, DWORD *TableNumber2)"""
)

SLM_Ctrl_WriteMP = _libraries[_libname].SLM_Ctrl_WriteMP
SLM_Ctrl_WriteMP.restype = SLM_STATUS
SLM_Ctrl_WriteMP.argtypes = [DWORD, DWORD]
SLM_Ctrl_WriteMP.__doc__ = """SLM_Ctrl_WriteMP(DWORD SLMNumber, DWORD TableNumber)"""

SLM_Ctrl_WriteMZ = _libraries[_libname].SLM_Ctrl_WriteMZ
SLM_Ctrl_WriteMZ.restype = SLM_STATUS
SLM_Ctrl_WriteMZ.argtypes = [DWORD]
SLM_Ctrl_WriteMZ.__doc__ = """SLM_Ctrl_WriteMZ(DWORD SLMNumber)"""

SLM_Ctrl_WriteMW = _libraries[_libname].SLM_Ctrl_WriteMW
SLM_Ctrl_WriteMW.restype = SLM_STATUS
SLM_Ctrl_WriteMW.argtypes = [DWORD, DWORD]
SLM_Ctrl_WriteMW.__doc__ = """SLM_Ctrl_WriteMW(DWORD SLMNumber, DWORD frames)"""

SLM_Ctrl_ReadMW = _libraries[_libname].SLM_Ctrl_ReadMW
SLM_Ctrl_ReadMW.restype = SLM_STATUS
SLM_Ctrl_ReadMW.argtypes = [DWORD, c_void_p]
SLM_Ctrl_ReadMW.__doc__ = """SLM_Ctrl_ReadMW(DWORD SLMNumber, DWORD *frames)"""

SLM_Ctrl_WriteDS = _libraries[_libname].SLM_Ctrl_WriteDS
SLM_Ctrl_WriteDS.restype = SLM_STATUS
SLM_Ctrl_WriteDS.argtypes = [DWORD, DWORD]
SLM_Ctrl_WriteDS.__doc__ = """SLM_Ctrl_WriteDS(DWORD SLMNumber, DWORD MemoryNumber)"""

SLM_Ctrl_ReadDS = _libraries[_libname].SLM_Ctrl_ReadDS
SLM_Ctrl_ReadDS.restype = SLM_STATUS
SLM_Ctrl_ReadDS.argtypes = [DWORD, c_void_p]
SLM_Ctrl_ReadDS.__doc__ = """SLM_Ctrl_ReadDS(DWORD SLMNumber, DWORD *MemoryNumber)"""

SLM_Ctrl_WriteDR = _libraries[_libname].SLM_Ctrl_WriteDR
SLM_Ctrl_WriteDR.restype = SLM_STATUS
SLM_Ctrl_WriteDR.argtypes = [DWORD, DWORD]
SLM_Ctrl_WriteDR.__doc__ = """SLM_Ctrl_WriteDR(DWORD SLMNumber, DWORD order)"""

SLM_Ctrl_WriteDB = _libraries[_libname].SLM_Ctrl_WriteDB
SLM_Ctrl_WriteDB.restype = SLM_STATUS
SLM_Ctrl_WriteDB.argtypes = [DWORD]
SLM_Ctrl_WriteDB.__doc__ = """SLM_Ctrl_WriteDB(DWORD SLMNumber)"""

SLM_Ctrl_WriteGS = _libraries[_libname].SLM_Ctrl_WriteGS
SLM_Ctrl_WriteGS.restype = SLM_STATUS
SLM_Ctrl_WriteGS.argtypes = [DWORD, c_void_p]
SLM_Ctrl_WriteGS.__doc__ = """SLM_Ctrl_WriteGS(DWORD SLMNumber, USHORT GrayScale)"""

SLM_Ctrl_ReadGS = _libraries[_libname].SLM_Ctrl_ReadGS
SLM_Ctrl_ReadGS.restype = SLM_STATUS
SLM_Ctrl_ReadGS.argtypes = [DWORD, LPUSHORT]
SLM_Ctrl_ReadGS.__doc__ = """SLM_Ctrl_ReadGS(DWORD SLMNumber, USHORT *GrayScale)"""

SLM_Ctrl_ReadT = _libraries[_libname].SLM_Ctrl_ReadT
SLM_Ctrl_ReadT.restype = SLM_STATUS
SLM_Ctrl_ReadT.argtypes = [DWORD, LPDWORD, LPDWORD]
SLM_Ctrl_ReadT.__doc__ = (
    """SLM_Ctrl_ReadT(DWORD SLMNumber, float *deviceTemp, float *optionTemp)"""
)

SLM_Ctrl_ReadEDO = _libraries[_libname].SLM_Ctrl_ReadEDO
SLM_Ctrl_ReadEDO.restype = SLM_STATUS
SLM_Ctrl_ReadEDO.argtypes = [DWORD, LPDWORD, LPDWORD]
SLM_Ctrl_ReadEDO.__doc__ = (
    """SLM_Ctrl_ReadEDO(DWORD SLMNumber, DWORD *deviceError, DWORD *optionError)"""
)

SLM_Ctrl_ReadSU = _libraries[_libname].SLM_Ctrl_ReadSU
SLM_Ctrl_ReadSU.restype = SLM_STATUS
SLM_Ctrl_ReadSU.argtypes = [DWORD]
SLM_Ctrl_ReadSU.__doc__ = """SLM_Ctrl_ReadSU(DWORD SLMNumber)"""

SLM_Ctrl_ReadSDO = _libraries[_libname].SLM_Ctrl_ReadSDO
SLM_Ctrl_ReadSDO.restype = SLM_STATUS
SLM_Ctrl_ReadSDO.argtypes = [DWORD, LPSTR, LPSTR]
SLM_Ctrl_ReadSDO.__doc__ = (
    """SLM_Ctrl_ReadSDO(DWORD SLMNumber, LPSTR deviceID, LPSTR optionID)"""
)



"""
Hardware control for Meadowlark SLMs.
Tested with Meadowlark (AVR Optics) P1920-400-800-HDMI-T.

Note
~~~~
Check that the Blink SDK, including DLL files etc, are in the default folder
or otherwise pass the correct directory in the constructor.
"""
import os
import ctypes
import warnings

from .slm import SLM

DEFAULT_SDK_PATH = "C:\\Program Files\\Meadowlark Optics\\Blink 1920 HDMI\\"

class Meadowlark(SLM):
    """
    Interfaces with Meadowlark SLMs.

    Attributes
    ----------
    slm_lib : ctypes.CDLL
        Connection to the Meadowlark library.
    sdk_path : str
        Path of the Blink SDK folder.
    """

    def __init__(
        self,
        verbose=True,
        sdk_path=DEFAULT_SDK_PATH,
        lut_path=None,
        wav_um=1,
        pitch_um=(8,8),
        **kwargs
    ):
        r"""
        Initializes an instance of a Meadowlark SLM.

        Caution
        ~~~~~~~
        :class:`.Meadowlark` defaults to 8 micron SLM pixel size.
        This is valid for most Meadowlark models, but not true for all!

        Arguments
        ---------
        verbose : bool
            Whether to print extra information.
        sdk_path : str
            Path of the Blink SDK installation folder. Stored in :attr:`sdk_path`.

            Important
            ~~~~~~~~~
            If the installation is not in the default folder,
            then this path needs to be corrected.
        lut_path : str OR None
            Passed to :meth:`load_lut`. Looks for the voltage 'look-up table' data
            which is necessary to run the SLM.

            Tip
            ~~~
            See :meth:`load_lut` for how the default
            argument and other options are parsed.
        wav_um : float
            Wavelength of operation in microns. Defaults to 1 um.
        pitch_um : (float, float)
            Pixel pitch in microns. Defaults to 8 micron square pixels.
        **kwargs
            See :meth:`.SLM.__init__` for permissible options.
        """
        # Validates the DPI awareness of this context, which is presumably important for scaling.
        if verbose: print("Validating DPI awareness...", end="")

        awareness = ctypes.c_int()
        error_get = ctypes.windll.shcore.GetProcessDpiAwareness(0, ctypes.byref(awareness))
        error_set = ctypes.windll.shcore.SetProcessDpiAwareness(2)
        success = ctypes.windll.user32.SetProcessDPIAware()

        if not success:
            raise RuntimeError(
                "Meadowlark failed to validate DPI awareness. "
                "Errors: get={}, set={}, awareness={}".format(error_get, error_set, awareness.value)
            )
        if verbose: print("success")

        # Open the SLM library
        if verbose: print("Constructing Blink SDK...", end="")

        dll_path = os.path.join(sdk_path, "SDK", "Blink_C_wrapper")
        try:
            ctypes.cdll.LoadLibrary(dll_path)
            self.slm_lib = ctypes.CDLL("Blink_C_wrapper")
        except:
            print("failure")
            raise ImportError(
                f"Meadowlark .dlls did not did not import correctly. "
                f"Is '{dll_path}' the correct path?"
            )

        self.sdk_path = sdk_path

        # Initialize the SDK. The requirements of Matlab, LabVIEW, C++ and Python are different, so pass
        # the constructor a boolean indicating if we are calling from C++/Python (true), or Matlab/LabVIEW (false)
        bool_cpp_or_python = ctypes.c_uint(1)
        self.slm_lib.Create_SDK(bool_cpp_or_python)

        # Adjust pre- and post-ramp slopes for accurate voltage setting
        # (otherwise, custom LUT calibration is not properly implemented [this feature is not implemented in slmsuite]).
        # You may need a special version of the SDK sent to you from Meadowlark to have access to these parameters.
        # self.slm_lib.SetPreRampSlope(20) # default is 7
        # self.slm_lib.SetPostRampSlope(24) # default is 24

        if verbose: print("success")

        # Load LUT.
        if verbose: print("Loading LUT file...", end="")

        try:
            true_lut_path = self.load_lut(lut_path)
        except RuntimeError as e:
            if verbose: print("failure\n(could not find .lut file)")
            raise e
        else:
            if verbose and true_lut_path != lut_path:
                print(f"success\n(loaded from '{true_lut_path}')")

        # Construct other variables.
        super().__init__(
            (self.slm_lib.Get_Width(), self.slm_lib.Get_Height()),
            bitdepth=self.slm_lib.Get_Depth(),
            name=kwargs.pop("name", "Meadowlark"),
            wav_um=wav_um,
            pitch_um=pitch_um,
            **kwargs
        )

        if self.bitdepth > 8:
            warnings.warn(
                f"Bitdepth of {self.bitdepth} > 8 detected; "
                "this has not been tested and might fail."
            )

        self.write(None)

    def load_lut(self, lut_path=None):
        """
        Loads a voltage 'look-up table' (LUT) to the SLM.
        This converts requested phase values to physical voltage perturbing
        the liquid crystals.

        Parameters
        ----------
        lut_path : str OR None
            Path to look for an LUT file in.

            -   If this is a .lut file, then this file is loaded to the SLM.
            -   If this is a directory, then searches all files inside the
                directory, and loads either the alphabetically-first .lut file
                or if possible the alphabetically-first .lut file starting with ``"slm"``
                which is more likely to correspond to the LUT customized to an SLM,
                as Meadowlark sends such files prefixed by
                ``"slm"`` such as ``"slm5758_at532.lut"``.

        Raises
        ------
        RuntimeError
            If a .lut file is not found.

        Returns
        -------
        str
            The path which was used to load the LUT.
        """
        # If a path is not given, search inside the SDK path.
        if lut_path is None:
            lut_path = os.path.join(self.sdk_path, "LUT Files")

        # If we already have a .lut file, proceed.
        if len(lut_path) > 4 and lut_path[-4:] == ".lut":
            pass
        else:   # Otherwise, treat the path like a folder and search inside the folder.
            lut_file = None

            for file in os.listdir(lut_path):
                # Only examine .lut files.
                if len(file) >= 4 and file[-4:].lower() == ".lut":
                    # Choose the first one.
                    if lut_file is None:
                        lut_file = file

                    # Or choose the first one that starts with "slm"
                    if file[:3].lower() == "slm" and not lut_file[:3].lower() == "slm":
                        lut_file = file
                        break

            # Throw an error if we didn't find a .lut file.
            if lut_file is not None:
                lut_path = os.path.join(lut_path, lut_file)
            else:
                raise RuntimeError(
                    f"Could not find a .lut file at path '{lut_path}'"
                )

        # Finally, load the lookup table.
        self.slm_lib.Load_lut(lut_path)

        return lut_path

    @staticmethod
    def info(verbose=True):
        """
        The normal behavior of this function is to discover the names of all the displays
        to help the user identify the correct display. However, Meadowlark software does
        not currently support multiple SLMs, so this function instead raises an error.

        Parameters
        ----------
        verbose : bool
            Whether to print the discovered information.

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError(
            "Meadowlark software does not currently support multiple SLMs, "
            "so a function to identify SLMs is moot. "
            "If functionality with multiple SLMs is desired, contact them directly."
        )

    def close(self):
        """
        See :meth:`.SLM.close`.
        """
        self.slm_lib.Delete_SDK()

    def _write_hw(self, display):
        """
        See :meth:`.SLM._write_hw`.
        """
        self.slm_lib.Write_image(
            display.ctypes.data_as(ctypes.POINTER(ctypes.c_ubyte)),
            ctypes.c_uint(self.bitdepth == 8)   # Is 8-bit
        )

    ### Additional Meadowlark-specific functionality

    def get_temperature(self):
        """
        Read the temperature of the SLM.

        Returns
        -------
        float
            Temperature in degrees celcius.
        """
        return self.slm_lib.Get_SLMTemp()

    def get_coverglass_voltage(self):
        """
        Read the voltage of the SLM coverglass.

        Returns
        -------
        float
            Voltage of the SLM coverglass.
        """
        return self.slm_lib.Get_SLMVCom()


"""
Abstract camera functionality.
"""
import time
import warnings
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy.optimize import curve_fit

from slmsuite.holography import analysis
from slmsuite.holography.toolbox import BLAZE_LABELS
from slmsuite.misc.fitfunctions import lorentzian, lorentzian_jacobian
from slmsuite.misc.math import REAL_TYPES


class Camera():
    """
    Abstract class for cameras.
    Comes with transformations, averaging and HDR,
    and helper functions like :meth:`.autoexpose()`.

    Attributes
    ----------
    name : str
        Camera identifier.
    shape : (int, int)
        Stores ``(height, width)`` of the camera in pixels, the same convention as
        :meth:`numpy.shape`.
    bitdepth : int
        Depth of a camera pixel well in bits.
    bitresolution : int
        Stores ``2**bitdepth``.
    dtype : type
        Stores the type returned by :meth:`._get_image_hw()`.
        This value is cached upon initialization.
    averaging : int OR None
        Default setting for averaging. See :meth:`.get_image()`.
    hdr : (int, int) OR None
        Default setting for multi-exposure High Dynamic Range imaging. See :meth:`.get_image()`.
    capture_attempts : int
        If the camera returns an error or exceeds a timeout, 
        try again for a total of `capture_attempts` attempts.
        This is useful for resilience against errors that happen with low probability.
        Defaults to 5.
    pitch_um : (float, float) OR None
        Pixel pitch in microns.
    exposure_bounds_s : (float, float) OR None
        Shortest and longest allowable integration in seconds.
    woi : tuple
        WOI (window of interest) in ``(x, width, y, height)`` form.

        Warning
        ~~~~~~~
        This feature is less fleshed out than most. There may be issues
        (e.g. :meth:`.get_image()` with the ``averaging`` or ``hdr`` flags).
    default_shape : tuple
        Default ``shape`` of the camera before any WOI or transform changes are made.
    transform : function
        Flip and/or rotation operator specified by the user in :meth:`__init__`.
        The user is expected to apply this transform to the matrix returned in
        :meth:`get_image()`. Note that WOI changes are applied on the camera hardware
        **before** this transformation.
    """

    def __init__(
        self,
        resolution,
        bitdepth=8,
        pitch_um=None,
        name="camera",
        averaging=None,
        capture_attempts=5,
        hdr=None,
        rot="0",
        fliplr=False,
        flipud=False,
    ):
        """
        Initializes a camera.

        In addition to the other class attributes, accepts the following parameters
        to set :attr:`transform`. See :meth:`~slmsuite.holography.analysis.get_orientation_transformation()`.

        Parameters
        ----------
        resolution
            The width and height of the camera in ``(width, height)`` form.

            Important
            ~~~~~~~~~
            This is the opposite of the numpy ``(height, width)``
            convention stored in :attr:`shape`.
        bitdepth
            See :attr:`bitdepth`.
        pitch_um : (float, float) OR None
            Fill in extra information about the pixel pitch in ``(dx_um, dy_um)`` form
            to use additional calibrations.
        name : str
            Defaults to ``"camera"``.
        averaging : int or None
            Number of frames to average. Used to increase the effective bit depth of a camera by using
            pre-quantization noise (e.g. dark current, read-noise, etc.) to "dither" the pixel output
            signal. If ``None``, no averaging is performed.
        hdr : int OR (int, int) OR None OR False
            Exposure information for `Multi-exposure High Dynamic Range (HDR) imaging
            <https://en.wikipedia.org/wiki/Multi-exposure_HDR_capture>`_
        capture_attempts : int
            If the camera returns an error or exceeds a timeout, 
            try again for a total of `capture_attempts` attempts.
            This is useful for resilience against errors that happen with low probability.
            Defaults to 5.
        rot : str or int
            Rotates returned image by the corresponding degrees in ``["90", "180", "270"]``
            or :meth:`numpy.rot90` code in ``[1, 2, 3]``. Defaults to no rotation.
            Used to determine :attr:`shape` and :attr:`transform`.
        fliplr : bool
            Flips returned image left right.
            Used to determine :attr:`transform`.
        flipud : bool
            Flips returned image up down.
            Used to determine :attr:`transform`.
        """
        (width, height) = resolution

        # Set shape, depending upon transform.
        if rot in ("90", 1, "270", 3):
            self.shape = (width, height)
            self.default_shape = (width, height)
        else:
            self.shape = (height, width)
            self.default_shape = (height, width)

        # Parse capture_attempts.
        self.capture_attempts = int(capture_attempts)
        if capture_attempts <= 0:
            raise ValueError("capture_attempts must be positive.")

        # Create image transformation.
        self.transform = analysis.get_orientation_transformation(rot, fliplr, flipud)

        # Update WOI information.
        self.woi = (0, width, 0, height)
        try:
            self.set_woi()
        except NotImplementedError:
            pass

        # Set other useful parameters
        self.name = str(name)
        self.bitdepth = int(bitdepth)
        self.bitresolution = 2**bitdepth
        self.dtype = self._get_dtype()

        # Frame averaging
        self.averaging = self._parse_averaging(averaging, preserve_none=True)
        self.hdr = self._parse_hdr(hdr, preserve_none=True)

        # Spatial dimensions
        if pitch_um is not None:
            if isinstance(pitch_um, REAL_TYPES):
                pitch_um = [pitch_um, pitch_um]
            self.pitch_um = np.squeeze(pitch_um)
            if (len(self.pitch_um) != 2):
                raise ValueError("Expected (float, float) for pitch_um")
            self.pitch_um = np.array([float(self.pitch_um[0]), float(self.pitch_um[1])])
        else:
            self.pitch_um = None

        # Default to None, allow subclass constructors to fill.
        self.exposure_bounds_s = None

    # Core methods - to be implemented by subclass.

    def close(self):
        """
        Abstract method to close the camera and delete related objects.
        """
        raise NotImplementedError()

    @staticmethod
    def info(verbose=True):
        """
        Abstract method to load display information.

        Parameters
        ----------
        verbose : bool
            Whether or not to print display information.

        Returns
        -------
        list
            An empty list.
        """
        if verbose:
            print(".info() NotImplemented.")
        return []

    def reset(self):
        """
        Abstract method to reset the camera to a default state.
        """
        raise NotImplementedError()

    def get_exposure(self):
        """
        Abstract method to get the integration time in seconds.
        Used in :meth:`.autoexposure()`.

        Returns
        -------
        float
            Integration time in seconds.
        """
        raise NotImplementedError()

    def set_exposure(self, exposure_s):
        """
        Abstract method to set the integration time in seconds.
        Used in :meth:`.autoexposure()`.

        Parameters
        ----------
        exposure_s : float
            The integration time in seconds.
        """
        raise NotImplementedError()

    def set_woi(self, woi=None):
        """
        Abstract method to narrow the imaging region to a 'window of interest'
        for faster framerates.

        Parameters
        ----------
        woi : list, None
            See :attr:`~slmsuite.hardware.cameras.camera.Camera.woi`.
            If ``None``, defaults to largest possible.

        Returns
        ----------
        woi : list
            :attr:`~slmsuite.hardware.cameras.camera.Camera.woi`.
        """
        raise NotImplementedError()

    def flush(self, timeout_s=1):
        """
        Abstract method to cycle the image buffer (if any)
        such that all new :meth:`.get_image()`
        calls yield fresh frames.

        Parameters
        ----------
        timeout_s : float
            The time in seconds to wait for frames to catch up with triggers.
        """
        raise NotImplementedError()

    def _get_image_hw(self, timeout_s=1):
        """
        Abstract method to capture camera images.

        Parameters
        ----------
        timeout_s : float
            The time in seconds to wait for the frame to be fetched.

        Returns
        -------
        numpy.ndarray
            Array of shape :attr:`~slmsuite.hardware.cameras.camera.Camera.shape`.
        """
        raise NotImplementedError()

    def _get_images_hw(self, image_count, timeout_s=1, out=None):
        """
        Abstract method to capture a series of image_count images using camera-specific
        batch acquisition features.

        Parameters
        ----------
        image_count : int
            Number of frames to batch collect.
        timeout_s : float
            The time in seconds to wait for the frame to be fetched.

        Returns
        -------
        numpy.ndarray
            Array of shape (n_frames, :attr:`~slmsuite.hardware.cameras.camera.Camera.shape`).
        """
        raise NotImplementedError()

    # Capture methods one level of abstraction above _get_image_hw().

    def _get_image_hw_tolerant(self, *args, **kwargs):
        err = None

        for i in range(self.capture_attempts):
            try:
                return self._get_image_hw(*args, **kwargs)
            except Exception as e:
                if i > 0: warnings.warn(f"'{self.name}' _get_image_hw() failed on attempt {i}.")
                err = e

        raise err

    def _get_images_hw_tolerant(self, *args, **kwargs):
        e = None
        
        for i in range(self.capture_attempts):
            try:
                return self._get_images_hw(*args, **kwargs)
            except Exception as e:
                if i > 0: warnings.warn(f"'{self.name}' _get_images_hw() failed on attempt {i}.")
                err = e

        raise err

    def _get_dtype(self):
        try:
            self.dtype = np.array(self._get_image_hw_tolerant()).dtype   # Future: check if cameras change this after init.
        except NotImplementedError:
            if self.bitdepth > 16:
                self.dtype = float
            elif self.bitdepth > 8:
                self.dtype = np.uint16
            else:
                self.dtype = np.uint8

        try:
            if self.dtype(0).nbytes * 8 < self.bitdepth:
                raise warnings.warn(
                    f"Camera '{self.name}' bitdepth of {self.bitdepth} does not conform "
                    f"with the image type {self.dtype} with {self.dtype.itemsize} bytes."
                )
        except:     # The above sometimes fails for non-numpy datatypes.
            pass

    def _parse_averaging(self, averaging=None, preserve_none=False):
        """
        Helper function to get a valid averaging.
        """
        if averaging is None:
            if preserve_none:
                return None
            if not hasattr(self, "averaging") or self.averaging is None:
                averaging = 1
            else:
                averaging = self.averaging
        elif averaging is False:
            averaging = 1
        averaging = int(averaging)

        if averaging <= 0:
            raise ValueError("Cannot have negative averaging.")

        return averaging

    def _parse_hdr(self, exposures=None, preserve_none=False):
        """
        Helper function to get a valid hdr parameters.
        """
        # Parse inputs
        if exposures is None:
            if preserve_none:
                return None
            if not hasattr(self, "hdr") or self.hdr is None:
                (exposures, exposure_power) = (1, 0)
            else:
                (exposures, exposure_power) = self._parse_hdr(self.hdr)
        elif exposures is False:
            exposures = 1
            exposure_power = 0
        elif np.isscalar(exposures):
            exposure_power = 2
        else:
            (exposures, exposure_power) = exposures

        # Force int so we have a chance of exposure aligning with camera clock.
        return (int(exposures), int(exposure_power))

    def get_averaging_dtype(self, averaging=None):
        """Returns the appropriate image datatype for ``averaging`` levels of averaging."""
        if averaging is None:
            averaging = self.averaging
        averaging = int(averaging)

        if averaging <= 0:
            raise ValueError("Cannot have negative averaging.")

        # Switch based on image type
        if self.dtype.kind == "i" or self.dtype.kind == "u":
            dtype_bitdepth = self.dtype.nbytes

            # Remove depth for signed integeter.
            if self.dtype.kind == "i":
                dtype_bitdepth -= 1

            extra_bits = int(np.rint(np.log2(averaging)))

            if self.bitdepth + extra_bits <= dtype_bitdepth:
                # If we can sustain the averaging with the current type, continue.
                return self.dtype
            else:
                # Otherwise, force floating point.
                return float
        elif self.dtype.kind == "f":
            # Return floating point.
            return self.dtype
        else:
            raise ValueError(f"Datatype {self.dtype} does not make sense as a camera return.")

    def get_image(self, timeout_s=1, transform=True, hdr=None, averaging=None):
        """
        Capture, process, and return images from a camera.

        Tip
        ~~~
        This function includes two advanced capture options:

        -   `Multi-exposure High Dynamic Range (HDR) imaging
            <https://en.wikipedia.org/wiki/Multi-exposure_HDR_capture>`_
            and
        -   Software frame averaging (integrating).

        These methods can aid the user in capturing more precise data, beyond the
        default raw (and bitdepth-limited) output of the camera.

        Parameters
        ----------
        timeout_s : float
            The time in seconds to wait for the frame to be fetched.
        transform : bool
            Whether or not to transform the output image according to
            :attr:`~slmsuite.hardware.cameras.camera.Camera.transform`.
            Defaults to ``True``.
        hdr : int OR (int, int) OR None OR False
            Exposure information for `Multi-exposure High Dynamic Range (HDR) imaging
            <https://en.wikipedia.org/wiki/Multi-exposure_HDR_capture>`_
            If ``None``, the value of :attr:`hdr` is used.
            If ``False``, HDR is not used no matter the state of :attr:`hdr`.

            See Also
            ~~~~~~~~
            :meth:`.get_image_hdr()` for more information.

        averaging : int OR None OR False
            If ``int``, the number of frames to average over.
            If ``None``, the value of :attr:`averaging` is used.
            If ``False``, averaging is not used no matter the state of :attr:`averaging`.

            Tip
            ~~~
            The datatype is promoted to float if necessary but otherwise tries to stick
            with the default datatype.
            For instance, a camera that returns a 12-bit image as a 16-bit type has four
            more bits to use for averaging, i.e. :math:`2^4 = 16` possible averages without
            risk of overflow.
            Requesting more than 16 averages would cause the return type to be promoted
            to ``float``.

            Note
            ~~~~
            Averaging is a bit of a misnomer as the true functionality is to sum or
            integrate the images. This is done such that integer datatypes (useful for
            memory compactness) can still be returned; a general mean would need to be
            floating point.

        Returns
        -------
        numpy.ndarray of int OR float
            Array of shape :attr:`~slmsuite.hardware.cameras.camera.Camera.shape`.
        """
        # Parse acquisition options.
        averaging = self._parse_averaging(averaging)
        (exposures, exposure_power) = self._parse_hdr(hdr)

        # Switch based on what imaging case we're in.
        if exposures > 1:       # Average many images with increasing exposure.
            return self.get_image_hdr(
                (exposures, exposure_power),
                timeout_s=timeout_s,
                transform=transform,
                averaging=averaging,
            )
        elif averaging > 1:     # Average many images.
            averaging_dtype = self.get_averaging_dtype(averaging)

            try:
                # Using the camera-specific batch method if available
                imgs = self._get_images_hw(
                    averaging, timeout_s=timeout_s
                ).astype(averaging_dtype)

                # Cast as the proper type so we can sum.
                img = np.sum(imgs, axis=0)
            except NotImplementedError:
                # Brute-force collection as a backup
                img = np.zeros(self.default_shape, dtype=averaging_dtype)

                for _ in range(averaging):
                    img += self._get_image_hw_tolerant(timeout_s=timeout_s).astype(averaging_dtype)
        else:                   # Normal image
            img = self._get_image_hw_tolerant(timeout_s=timeout_s)

        # self.transform implements the flipping and rotating keywords passed to the
        # superclass constructor.
        if transform:
            img = self.transform(img)

        return img

    def get_images(self, image_count, timeout_s=1, out=None, transform=True, flush=False):
        """
        Grab ``image_count`` images in succession.

        Important
        ~~~~~~~~~
        This method does not support averaging or HDR features.
        Rather, it just returns a series of raw images.

        Parameters
        ----------
        image_count : int
            Number of images to grab.
        timeout_s : float
            The time in seconds to wait **for each** frame to be fetched.
        out : None OR numpy.ndarray
            If not ``None``, output data in this memory. Useful to avoid excessive allocation.
        transform : bool
            Whether or not to transform the output image according to
            :attr:`~slmsuite.hardware.cameras.camera.Camera.transform`.
            Defaults to ``True``.
        flush : bool
            Whether to flush before grabbing.

        Returns
        -------
        numpy.ndarray
            Array of shape ``(image_count, height, width)``.
        """
        # Preallocate memory if necessary
        out_shape = (int(image_count), self.default_shape[0], self.default_shape[1])
        if out is None:
            imgs = np.empty(out_shape, dtype=self.dtype)
        else:
            if out.shape != out_shape:
                raise ValueError(f"Expected out to be of shape {out_shape}. Found {out.shape}.")
            if out.dtype != self.dtype:
                raise ValueError(f"Expected out to be of type {self.dtype}. Found {out.dtype}.")
            imgs = np.array(out, copy=False, dtype=self.dtype)

        # Flush if desired.
        if flush:
            self.flush()

        # Grab images (no transformation)
        try:
            # Using the camera-specific method if available
            imgs = self._get_images_hw(image_count, timeout_s=timeout_s, out=imgs)
        except NotImplementedError:
            # Brute-force collection as a backup
            for i in range(image_count):
                imgs[i, :, :] = self._get_image_hw_tolerant(timeout_s=timeout_s)

        # Transform if desired.
        if transform:
            imgs_ = np.empty(
                (int(image_count), self.shape[0], self.shape[1]),
                dtype=self.dtype
            )
            for i in range(image_count):
                imgs_[i, :, :] = self.transform(imgs[i])

            imgs = imgs_

        return imgs

    def get_image_hdr(self, exposures=None, return_raw=False, **kwargs):
        r"""
        Often, the necessities of precision applications exceed the bitdepth of a
        camera. One way to recover High Dynamic Range (HDR) imaging is to use
        `multiple exposures <https://en.wikipedia.org/wiki/Multi-exposure_HDR_capture>`_
        each with increasing exposure time. Then, these images can be stitched together
        as floating-point data, omitting data which is under- or over- exposed.

        Tip
        ~~~
        This feature can be accessed in :meth:`.get_image()`
        using :attr:`hdr` or the ``hdr=`` flag.
        This function is exposed here also to reveal the raw data using ``return_raw=``
        and to draw attention to this useful feature.

        Caution
        ~~~~~~~
        Camera exposure is sometimes poorly defined. This might cause incorrect
        assumptions of the exposure.
        In general, a larger base exposure will produce more accurate results as a
        greater number of sample clock periods are rounded to for smaller relative variation.
        Future modifications to :meth:`get_image_hdr_analysis()` might improve image stitching.

        Parameters
        ----------
        exposures : int OR (int, int)
            The number of exposures to take.
            Each exposure increases in time multiplicatively from the base value
            (original :meth:`get_exposure()`) by a factor :math:`p`.
            The :math:`i\text{th}` image has exposure time :math:`\tau \times p^i`, zero-indexed.
            The default base of :math:`p = 2` leads to ``exposures`` being equivalent to
            `spots <https://en.wikipedia.org/wiki/Exposure_value>`_.
            This base can be changed to another number by instead passing a tuple, where
            the second ``int`` defines the desired base.
        return_raw : bool
            If ``True``, returns the raw data (stack of images with count ``exposures``)
            instead of the processed data. The data can be processed using :meth:`get_image_hdr_analysis`
        **kwargs
            Passed to :meth:`.get_image()`.

        Returns
        -------
        numpy.ndarray of float
            Array of shape :attr:`~slmsuite.hardware.cameras.camera.Camera.shape`.

            Important
            ~~~~~~~~~
            The scale of the returned image is the same as the original exposure.
        """
        (exposures, exposure_power) = self._parse_hdr(exposures)

        # Make empty data and grab the original exposure time.
        original_exposure = self.get_exposure()
        imgs = np.empty((exposures, self.shape[0], self.shape[1]))

        for i in range(exposures):
            self.set_exposure(int(exposure_power ** i) * original_exposure)
            self.flush()    # Sometimes, cameras return bad frames after exposure change.
            imgs[i, :, :] = self.get_image(hdr=False, **kwargs)

        # Reset exposure.
        self.set_exposure(original_exposure)
        self.flush()

        if return_raw:
            return imgs
        else:
            img = self.get_image_hdr_analysis(imgs, exposure_power=exposure_power, overexposure_threshold=self.bitresolution/2)
            if np.max(img) >= self.bitresolution:
                warnings.warn("HDR image is overexposed.")
            return img

    @staticmethod
    def get_image_hdr_analysis(imgs, overexposure_threshold=None, exposure_power=2):
        r"""
        Analyzes raw data for High Dynamic Range (HDR) imaging
        `multiple exposures <https://en.wikipedia.org/wiki/Multi-exposure_HDR_capture>`_
        each with increasing exposure time.

        Parameters
        ----------
        imgs : array_like
            Stack of images with increasing exposure.
        overexposure_threshold : float OR None
            For each image (except the first), data is thrown out if values are above
            this threshold. If ``None``, the threshold defaults to half the maximum.
        exposure_power : int
            Each exposure increases in time multiplicatively from the base value
            (original :meth:`get_exposure()`) by this factor :math:`p`. The :math:`i\text{th}` image has
            exposure time :math:`\tau \times p^i`, zero-indexed.
            The default value of ``2`` leads to ``exposures`` being equivalent to
            `spots <https://en.wikipedia.org/wiki/Exposure_value>`_.

        Returns
        -------
        numpy.ndarray of float
            Array of shape :attr:`~slmsuite.hardware.cameras.camera.Camera.shape`.

            Important
            ~~~~~~~~~
            The scale of the returned image is the same as the original exposure.
        """
        # Parse arguments
        exposure_power = int(exposure_power)
        if overexposure_threshold is None:
            # Default to half exposure.
            overexposure_threshold = np.max(imgs) / 2

        img = None

        for i in range(imgs.shape[0]):
            img_current = imgs[i, :, :].astype(float)

            if i == 0:
                img = img_current
            else:
                # Overwrite data when greater precision is available.
                mask = img_current < overexposure_threshold
                img[mask] = img_current[mask] / float(exposure_power ** i)

        return img

    # Other helper methods.

    def plot(self, image=None, limits=None, title="Image", ax=None, cbar=True):
        """
        Plots the provided image.

        Parameters
        ----------
        image : ndarray OR None
            Image to be plotted. If ``None``, grabs an image from the camera.
        limits : None OR float OR [[float, float], [float, float]]
            Scales the limits by a given factor or uses the passed limits directly.
        title : str
            Title the axis.
        ax : matplotlib.pyplot.axis OR None
            Axis to plot upon.
        cbar : bool
            Also plot a colorbar.

        Returns
        -------
        matplotlib.pyplot.axis
            Axis of the plotted image.
        """
        if image is None:
            self.flush()
            image = self.get_image()
        image = np.array(image, copy=(False if np.__version__[0] == '1' else None))

        if len(plt.get_fignums()) > 0:
            fig = plt.gcf()
        else:
            fig = plt.figure(figsize=(20,8))

        if ax is not None:
            plt.sca(ax)

        im = plt.imshow(image)
        ax = plt.gca()

        if cbar:
            cax = make_axes_locatable(ax).append_axes("right", size="2%", pad=0.05)
            fig.colorbar(im, cax=cax, orientation="vertical")

        # ax.invert_yaxis()
        ax.set_title(title)

        if limits is not None and limits != 1:
            if np.isscalar(limits):
                axlim = [ax.get_xlim(), ax.get_ylim()]

                centers = np.mean(axlim, axis=1)
                deltas = np.squeeze(np.diff(axlim, axis=1)) * limits / 2

                limits = np.vstack((centers - deltas, centers + deltas)).T
            elif np.shape(limits) == (2,2):
                pass
            else:
                raise ValueError(f"limits format {limits} not recognized; provide a scalar or limits.")

            ax.set_xlim(limits[0])
            ax.set_ylim(limits[1])

        if image.shape == self.shape:
            ax.set_xlabel(BLAZE_LABELS["ij"][0])
            ax.set_ylabel(BLAZE_LABELS["ij"][1])

        plt.sca(ax)

        return ax

    def autoexposure(
        self,
        set_fraction=0.5,
        tol=0.05,
        exposure_bounds_s=None,
        window=None,
        average_count=5,
        timeout_s=5,
        verbose=True,
    ):
        """
        Sets the exposure of the camera such that the maximum value is at ``set_fraction``
        of the dynamic range. Useful for mitigating deleterious over- or under- exposure.

        Parameters
        --------
        set_fraction : float
            Fraction of camera dynamic range to target image maximum.
        tol : float
            Fractional tolerance for exposure adjustment.
        exposure_bounds_s : (float, float) OR None
            Shortest and longest allowable integration in seconds. If ``None``, defaults to
            :attr:`exposure_bounds_s`. If this attribute was not set (or not availible on
            a particular camera), then ``None`` instead defaults to unbounded.
        window : array_like or None
            See :attr:`~slmsuite.hardware.cameras.camera.Camera.window`.
            If ``None``, the full camera frame will be used.
        average_count : int
            Number of frames to average intensity over for noise reduction.
        timeout_s : float
            Stop attempting to autoexpose after ``timeout_s`` seconds.
        verbose : bool
            Whether to print exposure updates.

        Returns
        --------
        float
            Resulting exposure in seconds.
        """
        # Parse exposure_bounds_s
        if exposure_bounds_s is None:
            if self.exposure_bounds_s is None:
                exposure_bounds_s = (0, np.inf)
            else:
                exposure_bounds_s = self.exposure_bounds_s

        # Parse window
        if window is None:
            wxi = 0
            wxf = self.shape[1]
            wyi = 0
            wyf = self.shape[0]
        else:
            wxi = int(window[0] - window[1] / 2)
            wxf = int(window[0] + window[1] / 2)
            wyi = int(window[2] - window[3] / 2)
            wyf = int(window[2] + window[3] / 2)

        # Initialize loop
        set_val = 0.5 * self.bitresolution
        exp = self.get_exposure()
        im_mean = np.mean(self.get_images(average_count, flush=True), 0)
        im_max = np.amax(im_mean[wyi:wyf, wxi:wxf])

        # Calculate the error as a percent of the camera's bitresolution
        err = np.abs(im_max - set_val) / self.bitresolution
        t = time.perf_counter()

        # Loop until timeout expires or we meet tolerance
        while err > tol and time.perf_counter() - t < timeout_s:
            # Clip exposure steps to 0.5x -> 2x
            exp = exp / np.amax([0.5, np.amin([(im_max / set_val), 2])])
            exp = np.amax([exposure_bounds_s[0], np.amin([exp, exposure_bounds_s[1]])])
            self.set_exposure(exp)
            im_mean = np.mean(self.get_images(average_count, flush=True), 0)
            im_max = np.amax(im_mean[wyi:wyf, wxi:wxf])
            err = np.abs(im_max - set_val) / self.bitresolution

            if verbose:
                print("Reset exposure to %1.2fs; maximum image value = %d." % (exp, im_max))

        exp_fin = exp * 2 * set_fraction

        # The loop targets 50% of resolution
        if set_fraction != 0.5:  # Sets for full dynamic range
            self.set_exposure(exp_fin)

        return exp_fin

    def autofocus(self, get_z, set_z, z_list=None, plot=False):
        """
        Uses an FFT contrast metric to find optimal focus when scanning over some variable
        ``z``. This ``z`` often takes the form of a vertical stage to position a sample precisely
        at the plane of imaging of a lens or objective. The contrast metric works particularly
        well when combined with a projected spot array hologram.

        Parameters
        ----------
        get_z : function
            Gets the current position of the focusing stage. Should return a ``float``.
        set_z : function
            Sets the position of the focusing stage to a given ``float``.
        z_list : array_like or None
            ``z`` values to sweep over during search.
            Defaults (when ``None``) to ``numpy.linspace(-4,4,16)``.
        plot : bool
            Whether to provide illustrative plots.
        """
        if z_list is None:
            z_list = np.linspace(-4, 4, 16)

        self.flush()

        z_base = get_z()
        imlist = []
        z_list = z_list + z_base
        counts = np.zeros_like(z_list)

        set_z(z_list[0])

        for i, z in enumerate(z_list):
            print("Moving to " + str(z))
            set_z(z)

            # Take image.
            img = self.get_image()
            imlist.append(np.copy(img))

            # Evaluate metric.
            dft = np.fft.fftshift(np.fft.fft2(imlist[-1].astype(float)))
            dft_amp = np.abs(dft)
            dft_norm = dft_amp / np.amax(dft_amp)
            fom_ = np.sum(dft_norm)
            counts[i] = fom_
            if plot:
                _, axs = plt.subplots(1, 2)
                axs[0].imshow(imlist[-1])
                axs[0].set_title("Image")
                axs[0].set_xticks([])
                axs[0].set_yticks([])
                axs[1].imshow(dft_norm)
                axs[1].set_title(f"FFT\nFoM$ = \\int\\int $|FFT|$ / $max|FFT|$ = {fom_}$")
                axs[1].set_xticks([])
                axs[1].set_yticks([])
                plt.show()

        counts[0] = counts[1]

        popt0 = np.array(
            [z_list[np.argmax(counts)], np.max(counts) - np.min(counts), np.min(counts), 100]
        )

        try:
            popt, _ = curve_fit(
                lorentzian,
                z_list,
                counts,
                jac=lorentzian_jacobian,
                ftol=1e-5,
                p0=popt0,
            )
            z_opt = popt[0]
            c_opt = popt[1] + popt[2]
        except BaseException:
            print("Autofocus fit failed, using maximum fom as optimum...")
            z_opt = z_list[np.argmax(counts)]
            c_opt = counts[np.argmax(counts)]

        # Return to original state except focus z
        print("Moving to optimized value " + str(z_opt))
        set_z(z_opt)

        # Show result if desired
        if plot:
            plt.plot(z_list, counts)
            plt.xlabel(r"$z$ $\mu$m")
            plt.ylabel("fom: Data, Guess, & Fit")
            plt.title("Focus Sweep")
            plt.scatter(z_opt, c_opt)
            plt.plot(z_list, lorentzian(z_list, *popt0))
            lfit = None
            try:
                lfit = lorentzian(z_list, *popt)
            except BaseException:
                lfit = None
            if lfit is not None:
                plt.plot(z_list, lfit)
            plt.legend(["Data", "Guess", "Result"])
            plt.show()

            plt.imshow(self.get_image())
            plt.title("Focused Image")
            plt.show()

        return z_opt, imlist


def _view_continuous(cameras, cmap=None, facecolor=None, dpi=300):
    """
    Continuously get camera frames and plot them. Intended for use in jupyter notebooks.
    Activate ``%matplotlib notebook`` before calling this function. This method
    does not halt, exit with a keyboard interrupt.

    Important
    ~~~~~~~~~
    This is probably going to get replaced with a :mod:`pyglet` interface for viewing
    realtime camera outputs while cameras loaded into python.

    Parameters
    ----------
    cameras : list of :class:`Camera`
        The cameras to view continuously.
    cmap
        See :meth:`matplotlib.pyplot.imshow`.
    facecolor
        See :meth:`matplotlib.pyplot.figure`.
    dpi
        See :meth:`matplotlib.pyplot.figure`. Default is 300.
    """
    # Get camera information.
    cam_count = len(cameras)
    cams_max_height = cams_max_width = 0
    for cam_idx, cam in enumerate(cameras):
        cams_max_height = max(cams_max_height, cam.shape[0])
        cams_max_width = max(cams_max_width, cam.shape[1])

    # Create figure.
    plt.ion()
    figsize = np.array((cam_count * cams_max_width, cams_max_height)) * 2**-9
    fig, axs = plt.subplots(1, cam_count, figsize=figsize, facecolor=facecolor, dpi=dpi)
    axs = np.reshape(axs, cam_count)
    fig.tight_layout()
    fig.show()
    fig.canvas.draw()
    for cam_idx in range(cam_count):
        axs[cam_idx].tick_params(direction="in")

    # Plot continuously.
    while True:
        for cam_idx in range(cam_count):
            cam = cameras[cam_idx]
            ax = axs[cam_idx]
            img = cam.get_image()
            ax.clear()
            ax.imshow(img, interpolation=None, cmap=cmap)
        fig.canvas.draw()
        fig.canvas.flush_events()



"""
Template for writing a subclass for SLM hardware control in :mod:`slmsuite`.
Outlines which SLM superclass functions must be implemented.
"""
from .slm import SLM

class Template(SLM):
    """
    Template for implementing a new SLM subclass. Replace :class:`Template`
    with the desired subclass name. :class:`~slmsuite.hardware.slms.slm.SLM` is the
    superclass that sets the requirements for :class:`Template`.
    """

    def __init__(
        self,
        bitdepth=8,         # TODO: Remove these arguments if the SLM SDK
        wav_um=1,           #       has some function to read them in.
        pitch_um=(8,8),     #       Otherwise, the user must supply.
        **kwargs
    ):
        r"""
        Initialize SLM and attributes.

        Parameters
        ----------
        bitdepth : int
            Depth of SLM pixel well in bits. Defaults to 10.
        wav_um : float
            Wavelength of operation in microns. Defaults to 1 um.
        pitch_um : (float, float)
            Pixel pitch in microns. Defaults to 8 micron square pixels.
        **kwargs
            See :meth:`.SLM.__init__` for permissible options.

        Note
        ~~~~
        These arguments, which ultimately are used to instantiate the :class:`.SLM` superclass,
        may be more accurately filled by calling the SLM's SDK functions.
        See the other implemented SLM subclasses for examples.
        """

        # TODO: Insert code here to initialize the SLM hardware, load properties, etc.

        # Mandatory functions:
        # - Opening a connection to the device.

        sdk = TODO()

        # Other possibilities to consider:
        # - Setting the SLM's operating wavelength (wav_um).
        # - Updating the SLM's phase table if necessary, and/or setting the design
        #   wavelength (wav_design_um).
        # - Setting the SLM's default settle time (abstract class SLM uses
        #   settle_time_s=0.3 seconds). This is important for experimental feedback to
        #   allow the SLM to settle before viewing the result on a camera.
        # - Checking for and saving the SLM parameters (height, width, etc).

        # Instantiate the superclass
        super().__init__(
            (sdk.width(), sdk.height()),
            bitdepth=bitdepth,
            wav_um=wav_um,
            pitch_um=pitch_um,
            **kwargs
        )

        # Zero the display using the superclass `write()` function.
        self.write(None)

    @staticmethod
    def info(verbose=True):
        """
        Discovers all SLMs detected by an SDK.
        Useful for a user to identify the correct serial numbers / etc.

        Parameters
        ----------
        verbose : bool
            Whether to print the discovered information.

        Returns
        --------
        list of str
            List of serial numbers or identifiers.
        """
        raise NotImplementedError()
        serial_list = get_serial_list()     # TODO: Fill in proper function.
        return serial_list

    def _write_hw(self, phase):
        """
        Low-level hardware interface to write ``phase`` data onto the SLM.
        When the user calls the :meth:`.SLM.write` method of
        :class:`.SLM`, ``phase`` is error checked before calling
        :meth:`_write_hw()`. See :meth:`.SLM._write_hw` for further detail.
        """
        # TODO: Insert code here to write raw phase data to the SLM.